@@ -536,7 +536,10 @@ static void child_main(int child_num_arg)
 
     bucket_alloc = apr_bucket_alloc_create(pchild);
 
-    while (!die_now) {
+    /* die_now is set when AP_SIG_GRACEFUL is received in the child;
+     * shutdown_pending is set when SIGTERM is received when running
+     * in single process mode.  */
+    while (!die_now && !shutdown_pending) {
         conn_rec *current_conn;
         void *csd;
 
@@ -577,6 +580,12 @@ static void child_main(int child_num_arg)
                         if (one_process && shutdown_pending) {
                             return;
                         }
+                        else if (die_now) {
+                            /* In graceful stop/restart; drop the mutex
+                             * and terminate the child. */
+                            SAFE_ACCEPT(accept_mutex_off());
+                            clean_child_exit(0);
+                        }
                         continue;
                     }
                     /* Single Unix documents select as returning errnos
@@ -587,6 +596,7 @@ static void child_main(int child_num_arg)
                      */
                     ap_log_error(APLOG_MARK, APLOG_ERR, status,
                                  ap_server_conf, "apr_pollset_poll: (listen)");
+                    SAFE_ACCEPT(accept_mutex_off());
                     clean_child_exit(1);
                 }
 
