@@ -24,10 +24,15 @@ APR_DECLARE(apr_status_t) apr_proc_mutex_destroy(apr_proc_mutex_t *mutex)
     return apr_pool_cleanup_run(mutex->pool, mutex, apr_proc_mutex_cleanup);
 }
 
-static apr_status_t proc_mutex_no_tryacquire(apr_proc_mutex_t *new_mutex)
+#if APR_HAS_POSIXSEM_SERIALIZE || APR_HAS_FCNTL_SERIALIZE || \
+    APR_HAS_PROC_PTHREAD_SERIALIZE || APR_HAS_SYSVSEM_SERIALIZE
+static apr_status_t proc_mutex_no_child_init(apr_proc_mutex_t **mutex,
+                                             apr_pool_t *cont,
+                                             const char *fname)
 {
-    return APR_ENOTIMPL;
+    return APR_SUCCESS;
 }
+#endif    
 
 #if APR_HAS_POSIXSEM_SERIALIZE
 
@@ -35,28 +40,21 @@ static apr_status_t proc_mutex_no_tryacquire(apr_proc_mutex_t *new_mutex)
 #define SEM_FAILED (-1)
 #endif
 
-static void proc_mutex_posix_setup(void)
-{
-}
-
 static apr_status_t proc_mutex_posix_cleanup(void *mutex_)
 {
-    apr_proc_mutex_t *mutex=mutex_;
-    apr_status_t stat = APR_SUCCESS;
+    apr_proc_mutex_t *mutex = mutex_;
     
-    if (mutex->interproc->filedes != -1) {
-        if (sem_close((sem_t *)mutex->interproc->filedes) < 0) {
-            stat = errno;
-        }
+    if (sem_close(mutex->psem_interproc) < 0) {
+        return errno;
     }
-    return stat;
+
+    return APR_SUCCESS;
 }    
 
 static apr_status_t proc_mutex_posix_create(apr_proc_mutex_t *new_mutex,
                                             const char *fname)
 {
     sem_t *psem;
-    apr_status_t stat;
     char semname[31];
     apr_time_t now;
     unsigned long sec;
@@ -64,7 +62,6 @@ static apr_status_t proc_mutex_posix_create(apr_proc_mutex_t *new_mutex,
     
     new_mutex->interproc = apr_palloc(new_mutex->pool,
                                       sizeof(*new_mutex->interproc));
-    new_mutex->interproc->filedes = -1;
     /*
      * This bogusness is to follow what appears to be the
      * lowest common denominator in Posix semaphore naming:
@@ -82,33 +79,30 @@ static apr_status_t proc_mutex_posix_create(apr_proc_mutex_t *new_mutex,
      * implementation. Versions previous to Darwin 6.2 had the 14
      * char limit, but later rev's allow up to 31 characters.
      *
-     * FIXME: There is a small window of opportunity where
-     * instead of getting a new semaphore descriptor, we get
-     * a previously obtained one. This can happen if the requests
-     * are made at the "same time" and in the small span of time between
-     * the sem_open and the sem_unlink. Use of O_EXCL does not
-     * help here however...
-     *
      */
     now = apr_time_now();
     sec = apr_time_sec(now);
     usec = apr_time_usec(now);
     apr_snprintf(semname, sizeof(semname), "/ApR.%lxZ%lx", sec, usec);
-    psem = sem_open((const char *) semname, O_CREAT, 0644, 1);
-    if ((psem == (sem_t *)SEM_FAILED) && (errno == ENAMETOOLONG)) {
-        /* Oh well, good try */
-        semname[13] = '\0';
-        psem = sem_open((const char *) semname, O_CREAT, 0644, 1);
+    psem = sem_open(semname, O_CREAT | O_EXCL, 0644, 1);
+    if (psem == (sem_t *)SEM_FAILED) {
+        if (errno == ENAMETOOLONG) {
+            /* Oh well, good try */
+            semname[13] = '\0';
+        } else if (errno == EEXIST) {
+            apr_snprintf(semname, sizeof(semname), "/ApR.%lxZ%lx", usec, sec);
+        } else {
+            return errno;
+        }
+        psem = sem_open(semname, O_CREAT | O_EXCL, 0644, 1);
     }
 
     if (psem == (sem_t *)SEM_FAILED) {
-        stat = errno;
-        proc_mutex_posix_cleanup(new_mutex);
-        return stat;
+        return errno;
     }
     /* Ahhh. The joys of Posix sems. Predelete it... */
-    sem_unlink((const char *) semname);
-    new_mutex->interproc->filedes = (int)psem;	/* Ugg */
+    sem_unlink(semname);
+    new_mutex->psem_interproc = psem;
     new_mutex->fname = apr_pstrdup(new_mutex->pool, semname);
     apr_pool_cleanup_register(new_mutex->pool, (void *)new_mutex,
                               apr_proc_mutex_cleanup, 
@@ -118,9 +112,19 @@ static apr_status_t proc_mutex_posix_create(apr_proc_mutex_t *new_mutex,
 
 static apr_status_t proc_mutex_posix_acquire(apr_proc_mutex_t *mutex)
 {
-    int rc;
+    if (sem_wait(mutex->psem_interproc) < 0) {
+        return errno;
+    }
+    mutex->curr_locked = 1;
+    return APR_SUCCESS;
+}
 
-    if ((rc = sem_wait((sem_t *)mutex->interproc->filedes)) < 0) {
+static apr_status_t proc_mutex_posix_tryacquire(apr_proc_mutex_t *mutex)
+{
+    if (sem_trywait(mutex->psem_interproc) < 0) {
+        if (errno == EAGAIN) {
+            return APR_EBUSY;
+        }
         return errno;
     }
     mutex->curr_locked = 1;
@@ -129,23 +133,16 @@ static apr_status_t proc_mutex_posix_acquire(apr_proc_mutex_t *mutex)
 
 static apr_status_t proc_mutex_posix_release(apr_proc_mutex_t *mutex)
 {
-    int rc;
-
     mutex->curr_locked = 0;
-    if ((rc = sem_post((sem_t *)mutex->interproc->filedes)) < 0) {
+    if (sem_post(mutex->psem_interproc) < 0) {
+        /* any failure is probably fatal, so no big deal to leave
+         * ->curr_locked at 0. */
         return errno;
     }
     return APR_SUCCESS;
 }
 
-static apr_status_t proc_mutex_posix_child_init(apr_proc_mutex_t **mutex,
-                                                apr_pool_t *cont,
-                                                const char *fname)
-{
-    return APR_SUCCESS;
-}
-
-const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_posix_methods =
+static const apr_proc_mutex_unix_lock_methods_t mutex_posixsem_methods =
 {
 #if APR_PROCESS_LOCK_IS_GLOBAL || !APR_HAS_THREADS || defined(POSIXSEM_IS_GLOBAL)
     APR_PROCESS_LOCK_MECH_IS_GLOBAL,
@@ -154,10 +151,10 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_posix_methods =
 #endif
     proc_mutex_posix_create,
     proc_mutex_posix_acquire,
-    proc_mutex_no_tryacquire,
+    proc_mutex_posix_tryacquire,
     proc_mutex_posix_release,
     proc_mutex_posix_cleanup,
-    proc_mutex_posix_child_init,
+    proc_mutex_no_child_init,
     "posixsem"
 };
 
@@ -166,6 +163,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_posix_methods =
 #if APR_HAS_SYSVSEM_SERIALIZE
 
 static struct sembuf proc_mutex_op_on;
+static struct sembuf proc_mutex_op_try;
 static struct sembuf proc_mutex_op_off;
 
 static void proc_mutex_sysv_setup(void)
@@ -173,6 +171,9 @@ static void proc_mutex_sysv_setup(void)
     proc_mutex_op_on.sem_num = 0;
     proc_mutex_op_on.sem_op = -1;
     proc_mutex_op_on.sem_flg = SEM_UNDO;
+    proc_mutex_op_try.sem_num = 0;
+    proc_mutex_op_try.sem_op = -1;
+    proc_mutex_op_try.sem_flg = SEM_UNDO | IPC_NOWAIT;
     proc_mutex_op_off.sem_num = 0;
     proc_mutex_op_off.sem_op = 1;
     proc_mutex_op_off.sem_flg = SEM_UNDO;
@@ -231,6 +232,23 @@ static apr_status_t proc_mutex_sysv_acquire(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
+static apr_status_t proc_mutex_sysv_tryacquire(apr_proc_mutex_t *mutex)
+{
+    int rc;
+
+    do {
+        rc = semop(mutex->interproc->filedes, &proc_mutex_op_try, 1);
+    } while (rc < 0 && errno == EINTR);
+    if (rc < 0) {
+        if (errno == EAGAIN) {
+            return APR_EBUSY;
+        }
+        return errno;
+    }
+    mutex->curr_locked = 1;
+    return APR_SUCCESS;
+}
+
 static apr_status_t proc_mutex_sysv_release(apr_proc_mutex_t *mutex)
 {
     int rc;
@@ -245,12 +263,7 @@ static apr_status_t proc_mutex_sysv_release(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
-static apr_status_t proc_mutex_sysv_child_init(apr_proc_mutex_t **mutex, apr_pool_t *cont, const char *fname)
-{
-    return APR_SUCCESS;
-}
-
-const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_sysv_methods =
+static const apr_proc_mutex_unix_lock_methods_t mutex_sysv_methods =
 {
 #if APR_PROCESS_LOCK_IS_GLOBAL || !APR_HAS_THREADS || defined(SYSVSEM_IS_GLOBAL)
     APR_PROCESS_LOCK_MECH_IS_GLOBAL,
@@ -259,10 +272,10 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_sysv_methods =
 #endif
     proc_mutex_sysv_create,
     proc_mutex_sysv_acquire,
-    proc_mutex_no_tryacquire,
+    proc_mutex_sysv_tryacquire,
     proc_mutex_sysv_release,
     proc_mutex_sysv_cleanup,
-    proc_mutex_sysv_child_init,
+    proc_mutex_no_child_init,
     "sysvsem"
 };
 
@@ -270,10 +283,6 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_sysv_methods =
 
 #if APR_HAS_PROC_PTHREAD_SERIALIZE
 
-static void proc_mutex_proc_pthread_setup(void)
-{
-}
-
 static apr_status_t proc_mutex_proc_pthread_cleanup(void *mutex_)
 {
     apr_proc_mutex_t *mutex=mutex_;
@@ -281,17 +290,26 @@ static apr_status_t proc_mutex_proc_pthread_cleanup(void *mutex_)
 
     if (mutex->curr_locked == 1) {
         if ((rv = pthread_mutex_unlock(mutex->pthread_interproc))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
             rv = errno;
 #endif
             return rv;
-        } 
-        if (munmap((caddr_t)mutex->pthread_interproc, sizeof(pthread_mutex_t))){
-            return errno;
         }
     }
+    /* curr_locked is set to -1 until the mutex has been created */
+    if (mutex->curr_locked != -1) {
+        if ((rv = pthread_mutex_destroy(mutex->pthread_interproc))) {
+#ifdef HAVE_ZOS_PTHREADS
+            rv = errno;
+#endif
+            return rv;
+        }
+    }
+    if (munmap((caddr_t)mutex->pthread_interproc, sizeof(pthread_mutex_t))) {
+        return errno;
+    }
     return APR_SUCCESS;
-}    
+}
 
 static apr_status_t proc_mutex_proc_pthread_create(apr_proc_mutex_t *new_mutex,
                                                    const char *fname)
@@ -311,59 +329,68 @@ static apr_status_t proc_mutex_proc_pthread_create(apr_proc_mutex_t *new_mutex,
                                        PROT_READ | PROT_WRITE, MAP_SHARED,
                                        fd, 0); 
     if (new_mutex->pthread_interproc == (pthread_mutex_t *) (caddr_t) -1) {
+        close(fd);
         return errno;
     }
     close(fd);
+
+    new_mutex->curr_locked = -1; /* until the mutex has been created */
+
     if ((rv = pthread_mutexattr_init(&mattr))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         proc_mutex_proc_pthread_cleanup(new_mutex);
         return rv;
     }
     if ((rv = pthread_mutexattr_setpshared(&mattr, PTHREAD_PROCESS_SHARED))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         proc_mutex_proc_pthread_cleanup(new_mutex);
+        pthread_mutexattr_destroy(&mattr);
         return rv;
     }
 
 #ifdef HAVE_PTHREAD_MUTEX_ROBUST
     if ((rv = pthread_mutexattr_setrobust_np(&mattr, 
                                                PTHREAD_MUTEX_ROBUST_NP))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         proc_mutex_proc_pthread_cleanup(new_mutex);
+        pthread_mutexattr_destroy(&mattr);
         return rv;
     }
     if ((rv = pthread_mutexattr_setprotocol(&mattr, PTHREAD_PRIO_INHERIT))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         proc_mutex_proc_pthread_cleanup(new_mutex);
+        pthread_mutexattr_destroy(&mattr);
         return rv;
     }
 #endif /* HAVE_PTHREAD_MUTEX_ROBUST */
 
     if ((rv = pthread_mutex_init(new_mutex->pthread_interproc, &mattr))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         proc_mutex_proc_pthread_cleanup(new_mutex);
+        pthread_mutexattr_destroy(&mattr);
         return rv;
     }
 
+    new_mutex->curr_locked = 0; /* mutex created now */
+
     if ((rv = pthread_mutexattr_destroy(&mattr))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         proc_mutex_proc_pthread_cleanup(new_mutex);
         return rv;
     }
 
-    new_mutex->curr_locked = 0;
     apr_pool_cleanup_register(new_mutex->pool,
                               (void *)new_mutex,
                               apr_proc_mutex_cleanup, 
@@ -376,7 +403,7 @@ static apr_status_t proc_mutex_proc_pthread_acquire(apr_proc_mutex_t *mutex)
     apr_status_t rv;
 
     if ((rv = pthread_mutex_lock(mutex->pthread_interproc))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
 #ifdef HAVE_PTHREAD_MUTEX_ROBUST
@@ -394,7 +421,32 @@ static apr_status_t proc_mutex_proc_pthread_acquire(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
-/* TODO: Add proc_mutex_proc_pthread_tryacquire(apr_proc_mutex_t *mutex) */
+static apr_status_t proc_mutex_proc_pthread_tryacquire(apr_proc_mutex_t *mutex)
+{
+    apr_status_t rv;
+ 
+    if ((rv = pthread_mutex_trylock(mutex->pthread_interproc))) {
+#ifdef HAVE_ZOS_PTHREADS 
+        rv = errno;
+#endif
+        if (rv == EBUSY) {
+            return APR_EBUSY;
+        }
+#ifdef HAVE_PTHREAD_MUTEX_ROBUST
+        /* Okay, our owner died.  Let's try to make it consistent again. */
+        if (rv == EOWNERDEAD) {
+            pthread_mutex_consistent_np(mutex->pthread_interproc);
+            rv = APR_SUCCESS;
+        }
+        else
+            return rv;
+#else
+        return rv;
+#endif
+    }
+    mutex->curr_locked = 1;
+    return rv;
+}
 
 static apr_status_t proc_mutex_proc_pthread_release(apr_proc_mutex_t *mutex)
 {
@@ -402,7 +454,7 @@ static apr_status_t proc_mutex_proc_pthread_release(apr_proc_mutex_t *mutex)
 
     mutex->curr_locked = 0;
     if ((rv = pthread_mutex_unlock(mutex->pthread_interproc))) {
-#ifdef PTHREAD_SETS_ERRNO
+#ifdef HAVE_ZOS_PTHREADS
         rv = errno;
 #endif
         return rv;
@@ -410,22 +462,15 @@ static apr_status_t proc_mutex_proc_pthread_release(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
-static apr_status_t proc_mutex_proc_pthread_child_init(apr_proc_mutex_t **mutex,
-                                            apr_pool_t *cont, 
-                                            const char *fname)
-{
-    return APR_SUCCESS;
-}
-
-const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_proc_pthread_methods =
+static const apr_proc_mutex_unix_lock_methods_t mutex_proc_pthread_methods =
 {
     APR_PROCESS_LOCK_MECH_IS_GLOBAL,
     proc_mutex_proc_pthread_create,
     proc_mutex_proc_pthread_acquire,
-    proc_mutex_no_tryacquire,
+    proc_mutex_proc_pthread_tryacquire,
     proc_mutex_proc_pthread_release,
     proc_mutex_proc_pthread_cleanup,
-    proc_mutex_proc_pthread_child_init,
+    proc_mutex_no_child_init,
     "pthread"
 };
 
@@ -462,10 +507,8 @@ static apr_status_t proc_mutex_fcntl_cleanup(void *mutex_)
         if (status != APR_SUCCESS)
             return status;
     }
-    if (mutex->interproc) { /* if it was opened successfully */
-        apr_file_close(mutex->interproc);
-    }
-    return APR_SUCCESS;
+        
+    return apr_file_close(mutex->interproc);
 }    
 
 static apr_status_t proc_mutex_fcntl_create(apr_proc_mutex_t *new_mutex,
@@ -488,7 +531,6 @@ static apr_status_t proc_mutex_fcntl_create(apr_proc_mutex_t *new_mutex,
     }
  
     if (rv != APR_SUCCESS) {
-        proc_mutex_fcntl_cleanup(new_mutex);
         return rv;
     }
 
@@ -515,6 +557,27 @@ static apr_status_t proc_mutex_fcntl_acquire(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
+static apr_status_t proc_mutex_fcntl_tryacquire(apr_proc_mutex_t *mutex)
+{
+    int rc;
+
+    do {
+        rc = fcntl(mutex->interproc->filedes, F_SETLK, &proc_mutex_lock_it);
+    } while (rc < 0 && errno == EINTR);
+    if (rc < 0) {
+#if FCNTL_TRYACQUIRE_EACCES
+        if (errno == EACCES) {
+#else
+        if (errno == EAGAIN) {
+#endif
+            return APR_EBUSY;
+        }
+        return errno;
+    }
+    mutex->curr_locked = 1;
+    return APR_SUCCESS;
+}
+
 static apr_status_t proc_mutex_fcntl_release(apr_proc_mutex_t *mutex)
 {
     int rc;
@@ -529,14 +592,7 @@ static apr_status_t proc_mutex_fcntl_release(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
-static apr_status_t proc_mutex_fcntl_child_init(apr_proc_mutex_t **mutex,
-                                                apr_pool_t *pool, 
-                                                const char *fname)
-{
-    return APR_SUCCESS;
-}
-
-const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_fcntl_methods =
+static const apr_proc_mutex_unix_lock_methods_t mutex_fcntl_methods =
 {
 #if APR_PROCESS_LOCK_IS_GLOBAL || !APR_HAS_THREADS || defined(FCNTL_IS_GLOBAL)
     APR_PROCESS_LOCK_MECH_IS_GLOBAL,
@@ -545,10 +601,10 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_fcntl_methods =
 #endif
     proc_mutex_fcntl_create,
     proc_mutex_fcntl_acquire,
-    proc_mutex_no_tryacquire,
+    proc_mutex_fcntl_tryacquire,
     proc_mutex_fcntl_release,
     proc_mutex_fcntl_cleanup,
-    proc_mutex_fcntl_child_init,
+    proc_mutex_no_child_init,
     "fcntl"
 };
 
@@ -558,10 +614,6 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_fcntl_methods =
 
 static apr_status_t proc_mutex_flock_release(apr_proc_mutex_t *);
 
-static void proc_mutex_flock_setup(void)
-{
-}
-
 static apr_status_t proc_mutex_flock_cleanup(void *mutex_)
 {
     apr_status_t status;
@@ -623,6 +675,23 @@ static apr_status_t proc_mutex_flock_acquire(apr_proc_mutex_t *mutex)
     return APR_SUCCESS;
 }
 
+static apr_status_t proc_mutex_flock_tryacquire(apr_proc_mutex_t *mutex)
+{
+    int rc;
+
+    do {
+        rc = flock(mutex->interproc->filedes, LOCK_EX | LOCK_NB);
+    } while (rc < 0 && errno == EINTR);
+    if (rc < 0) {
+        if (errno == EWOULDBLOCK) {
+            return APR_EBUSY;
+        }
+        return errno;
+    }
+    mutex->curr_locked = 1;
+    return APR_SUCCESS;
+}
+
 static apr_status_t proc_mutex_flock_release(apr_proc_mutex_t *mutex)
 {
     int rc;
@@ -655,14 +724,13 @@ static apr_status_t proc_mutex_flock_child_init(apr_proc_mutex_t **mutex,
     rv = apr_file_open(&new_mutex->interproc, new_mutex->fname,
                        APR_WRITE, 0, new_mutex->pool);
     if (rv != APR_SUCCESS) {
-        proc_mutex_flock_cleanup(new_mutex);
         return rv;
     }
     *mutex = new_mutex;
     return APR_SUCCESS;
 }
 
-const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_flock_methods =
+static const apr_proc_mutex_unix_lock_methods_t mutex_flock_methods =
 {
 #if APR_PROCESS_LOCK_IS_GLOBAL || !APR_HAS_THREADS || defined(FLOCK_IS_GLOBAL)
     APR_PROCESS_LOCK_MECH_IS_GLOBAL,
@@ -671,7 +739,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_flock_methods =
 #endif
     proc_mutex_flock_create,
     proc_mutex_flock_acquire,
-    proc_mutex_no_tryacquire,
+    proc_mutex_flock_tryacquire,
     proc_mutex_flock_release,
     proc_mutex_flock_cleanup,
     proc_mutex_flock_child_init,
@@ -682,21 +750,13 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_flock_methods =
 
 void apr_proc_mutex_unix_setup_lock(void)
 {
-#if APR_HAS_POSIXSEM_SERIALIZE
-    proc_mutex_posix_setup();
-#endif
+    /* setup only needed for sysvsem and fnctl */
 #if APR_HAS_SYSVSEM_SERIALIZE
     proc_mutex_sysv_setup();
 #endif
-#if APR_HAS_PROC_PTHREAD_SERIALIZE
-    proc_mutex_proc_pthread_setup();
-#endif
 #if APR_HAS_FCNTL_SERIALIZE
     proc_mutex_fcntl_setup();
 #endif
-#if APR_HAS_FLOCK_SERIALIZE
-    proc_mutex_flock_setup();
-#endif
 }
 
 static apr_status_t proc_mutex_choose_method(apr_proc_mutex_t *new_mutex, apr_lockmech_e mech)
@@ -704,50 +764,50 @@ static apr_status_t proc_mutex_choose_method(apr_proc_mutex_t *new_mutex, apr_lo
     switch (mech) {
     case APR_LOCK_FCNTL:
 #if APR_HAS_FCNTL_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_fcntl_methods;
+        new_mutex->inter_meth = &mutex_fcntl_methods;
 #else
         return APR_ENOTIMPL;
 #endif
         break;
     case APR_LOCK_FLOCK:
 #if APR_HAS_FLOCK_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_flock_methods;
+        new_mutex->inter_meth = &mutex_flock_methods;
 #else
         return APR_ENOTIMPL;
 #endif
         break;
     case APR_LOCK_SYSVSEM:
 #if APR_HAS_SYSVSEM_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_sysv_methods;
+        new_mutex->inter_meth = &mutex_sysv_methods;
 #else
         return APR_ENOTIMPL;
 #endif
         break;
     case APR_LOCK_POSIXSEM:
 #if APR_HAS_POSIXSEM_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_posix_methods;
+        new_mutex->inter_meth = &mutex_posixsem_methods;
 #else
         return APR_ENOTIMPL;
 #endif
         break;
     case APR_LOCK_PROC_PTHREAD:
 #if APR_HAS_PROC_PTHREAD_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_proc_pthread_methods;
+        new_mutex->inter_meth = &mutex_proc_pthread_methods;
 #else
         return APR_ENOTIMPL;
 #endif
         break;
     case APR_LOCK_DEFAULT:
 #if APR_USE_FLOCK_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_flock_methods;
+        new_mutex->inter_meth = &mutex_flock_methods;
 #elif APR_USE_SYSVSEM_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_sysv_methods;
+        new_mutex->inter_meth = &mutex_sysv_methods;
 #elif APR_USE_FCNTL_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_fcntl_methods;
+        new_mutex->inter_meth = &mutex_fcntl_methods;
 #elif APR_USE_PROC_PTHREAD_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_proc_pthread_methods;
+        new_mutex->inter_meth = &mutex_proc_pthread_methods;
 #elif APR_USE_POSIXSEM_SERIALIZE
-        new_mutex->inter_meth = &apr_proc_mutex_unix_posix_methods;
+        new_mutex->inter_meth = &mutex_posixsem_methods;
 #else
         return APR_ENOTIMPL;
 #endif
@@ -796,13 +856,8 @@ APR_DECLARE(apr_status_t) apr_proc_mutex_create(apr_proc_mutex_t **mutex,
     apr_proc_mutex_t *new_mutex;
     apr_status_t rv;
 
-    new_mutex = (apr_proc_mutex_t *)apr_pcalloc(pool,
-                                                sizeof(apr_proc_mutex_t));
-
-    new_mutex->pool  = pool;
-#if APR_HAS_SYSVSEM_SERIALIZE || APR_HAS_FCNTL_SERIALIZE || APR_HAS_FLOCK_SERIALIZE || APR_HAS_POSIXSEM_SERIALIZE
-    new_mutex->interproc = NULL;
-#endif
+    new_mutex = apr_pcalloc(pool, sizeof(apr_proc_mutex_t));
+    new_mutex->pool = pool;
 
     if ((rv = proc_mutex_create(new_mutex, mech, fname)) != APR_SUCCESS)
         return rv;
@@ -845,13 +900,18 @@ APR_DECLARE(const char *) apr_proc_mutex_name(apr_proc_mutex_t *mutex)
 
 APR_DECLARE(const char *) apr_proc_mutex_lockfile(apr_proc_mutex_t *mutex)
 {
-    /* posix sems use the fname field but don't use a file,
-     * so be careful 
-     */
-    if (!strcmp(mutex->meth->name, "flock") ||
-        !strcmp(mutex->meth->name, "fcntl")) {
+    /* POSIX sems use the fname field but don't use a file,
+     * so be careful. */
+#if APR_HAS_FLOCK_SERIALIZE
+    if (mutex->meth == &mutex_flock_methods) {
         return mutex->fname;
     }
+#endif
+#if APR_HAS_FCNTL_SERIALIZE
+    if (mutex->meth == &mutex_fcntl_methods) {
+        return mutex->fname;
+    }
+#endif
     return NULL;
 }
 
